@@ -1,17 +1,13 @@
 # ext/mutable.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
 
 r"""Provide support for tracking of in-place changes to scalar values,
 which are propagated into ORM change events on owning parent objects.
 
-.. versionadded:: 0.7 :mod:`sqlalchemy.ext.mutable` replaces SQLAlchemy's
-   legacy approach to in-place mutations of scalar values; see
-   :ref:`07_migration_mutation_extension`.
-
 .. _mutable_scalars:
 
 Establishing Mutability on Scalar Column Values
@@ -43,7 +39,7 @@ JSON strings before being persisted::
 The usage of ``json`` is only for the purposes of example. The
 :mod:`sqlalchemy.ext.mutable` extension can be used
 with any type whose target Python type may be mutable, including
-:class:`.PickleType`, :class:`.postgresql.ARRAY`, etc.
+:class:`.PickleType`, :class:`_postgresql.ARRAY`, etc.
 
 When using the :mod:`sqlalchemy.ext.mutable` extension, the value itself
 tracks all parents which reference it.  Below, we illustrate a simple
@@ -236,14 +232,6 @@ be assigned an object value which represents information "composed" from one
 or more columns from the underlying mapped table. The usual example is that of
 a geometric "point", and is introduced in :ref:`mapper_composite`.
 
-.. versionchanged:: 0.7
-    The internals of :func:`.orm.composite` have been
-    greatly simplified and in-place mutation detection is no longer enabled by
-    default; instead, the user-defined value must detect changes on its own and
-    propagate them to all owning parents. The :mod:`sqlalchemy.ext.mutable`
-    extension provides the helper class :class:`.MutableComposite`, which is a
-    slight variant on the :class:`.Mutable` class.
-
 As is the case with :class:`.Mutable`, the user-defined composite class
 subclasses :class:`.MutableComposite` as a mixin, and detects and delivers
 change events to its parents via the :meth:`.MutableComposite.changed` method.
@@ -282,7 +270,7 @@ and to also route attribute set events via ``__setattr__`` to the
             return not self.__eq__(other)
 
 The :class:`.MutableComposite` class uses a Python metaclass to automatically
-establish listeners for any usage of :func:`.orm.composite` that specifies our
+establish listeners for any usage of :func:`_orm.composite` that specifies our
 ``Point`` type. Below, when ``Point`` is mapped to the ``Vertex`` class,
 listeners are established which will route change events from ``Point``
 objects to each of the ``Vertex.start`` and ``Vertex.end`` attributes::
@@ -341,10 +329,6 @@ make use of the custom composite type::
                 raise ValueError("tuple or Point expected")
             return value
 
-.. versionadded:: 0.7.10,0.8.0b2
-    Support for the :meth:`.MutableBase.coerce` method in conjunction with
-    objects of type :class:`.MutableComposite`.
-
 Supporting Pickling
 --------------------
 
@@ -370,13 +354,18 @@ pickling process of the parent's object-relational state so that the
 :meth:`MutableBase._parents` collection is restored to all ``Point`` objects.
 
 """
-from ..orm.attributes import flag_modified
-from .. import event, types
-from ..orm import mapper, object_mapper, Mapper
-from ..util import memoized_property
-from ..sql.base import SchemaEventTarget
+from collections import defaultdict
 import weakref
 
+from .. import event
+from .. import inspect
+from .. import types
+from ..orm import Mapper
+from ..orm import mapper
+from ..orm.attributes import flag_modified
+from ..sql.base import SchemaEventTarget
+from ..util import memoized_property
+
 
 class MutableBase(object):
     """Common base class to :class:`.Mutable`
@@ -386,12 +375,17 @@ class MutableBase(object):
 
     @memoized_property
     def _parents(self):
-        """Dictionary of parent object->attribute name on the parent.
+        """Dictionary of parent object's :class:`.InstanceState`->attribute
+        name on the parent.
 
         This attribute is a so-called "memoized" property.  It initializes
         itself with a new ``weakref.WeakKeyDictionary`` the first time
         it is accessed, returning the same object upon subsequent access.
 
+        .. versionchanged:: 1.4 the :class:`.InstanceState` is now used
+           as the key in the weak dictionary rather than the instance
+           itself.
+
         """
 
         return weakref.WeakKeyDictionary()
@@ -473,13 +467,13 @@ class MutableBase(object):
                 if coerce:
                     val = cls.coerce(key, val)
                     state.dict[key] = val
-                val._parents[state.obj()] = key
+                val._parents[state] = key
 
         def load_attrs(state, ctx, attrs):
             if not attrs or listen_keys.intersection(attrs):
                 load(state)
 
-        def set(target, value, oldvalue, initiator):
+        def set_(target, value, oldvalue, initiator):
             """Listen for set/replace events on the target
             data member.
 
@@ -494,35 +488,51 @@ class MutableBase(object):
             if not isinstance(value, cls):
                 value = cls.coerce(key, value)
             if value is not None:
-                value._parents[target.obj()] = key
+                value._parents[target] = key
             if isinstance(oldvalue, cls):
-                oldvalue._parents.pop(target.obj(), None)
+                oldvalue._parents.pop(inspect(target), None)
             return value
 
         def pickle(state, state_dict):
             val = state.dict.get(key, None)
             if val is not None:
-                if 'ext.mutable.values' not in state_dict:
-                    state_dict['ext.mutable.values'] = []
-                state_dict['ext.mutable.values'].append(val)
+                if "ext.mutable.values" not in state_dict:
+                    state_dict["ext.mutable.values"] = defaultdict(list)
+                state_dict["ext.mutable.values"][key].append(val)
 
         def unpickle(state, state_dict):
-            if 'ext.mutable.values' in state_dict:
-                for val in state_dict['ext.mutable.values']:
-                    val._parents[state.obj()] = key
+            if "ext.mutable.values" in state_dict:
+                collection = state_dict["ext.mutable.values"]
+                if isinstance(collection, list):
+                    # legacy format
+                    for val in collection:
+                        val._parents[state] = key
+                else:
+                    for val in state_dict["ext.mutable.values"][key]:
+                        val._parents[state] = key
 
-        event.listen(parent_cls, 'load', load,
-                     raw=True, propagate=True)
-        event.listen(parent_cls, 'refresh', load_attrs,
-                     raw=True, propagate=True)
-        event.listen(parent_cls, 'refresh_flush', load_attrs,
-                     raw=True, propagate=True)
-        event.listen(attribute, 'set', set,
-                     raw=True, retval=True, propagate=True)
-        event.listen(parent_cls, 'pickle', pickle,
-                     raw=True, propagate=True)
-        event.listen(parent_cls, 'unpickle', unpickle,
-                     raw=True, propagate=True)
+        event.listen(
+            parent_cls,
+            "_sa_event_merge_wo_load",
+            load,
+            raw=True,
+            propagate=True,
+        )
+
+        event.listen(parent_cls, "load", load, raw=True, propagate=True)
+        event.listen(
+            parent_cls, "refresh", load_attrs, raw=True, propagate=True
+        )
+        event.listen(
+            parent_cls, "refresh_flush", load_attrs, raw=True, propagate=True
+        )
+        event.listen(
+            attribute, "set", set_, raw=True, retval=True, propagate=True
+        )
+        event.listen(parent_cls, "pickle", pickle, raw=True, propagate=True)
+        event.listen(
+            parent_cls, "unpickle", unpickle, raw=True, propagate=True
+        )
 
 
 class Mutable(MutableBase):
@@ -537,7 +547,7 @@ class Mutable(MutableBase):
         """Subclasses should call this method whenever change events occur."""
 
         for parent, key in self._parents.items():
-            flag_modified(parent, key)
+            flag_modified(parent.obj(), key)
 
     @classmethod
     def associate_with_attribute(cls, attribute):
@@ -572,7 +582,7 @@ class Mutable(MutableBase):
                 if isinstance(prop.columns[0].type, sqltype):
                     cls.associate_with_attribute(getattr(class_, prop.key))
 
-        event.listen(mapper, 'mapper_configured', listen_for_type)
+        event.listen(mapper, "mapper_configured", listen_for_type)
 
     @classmethod
     def as_mutable(cls, sqltype):
@@ -613,9 +623,11 @@ class Mutable(MutableBase):
         # and we'll lose our ability to link that type back to the original.
         # so track our original type w/ columns
         if isinstance(sqltype, SchemaEventTarget):
+
             @event.listens_for(sqltype, "before_parent_attach")
             def _add_column_memo(sqltyp, parent):
-                parent.info['_ext_mutable_orig_type'] = sqltyp
+                parent.info["_ext_mutable_orig_type"] = sqltyp
+
             schema_event_check = True
         else:
             schema_event_check = False
@@ -625,16 +637,14 @@ class Mutable(MutableBase):
                 return
             for prop in mapper.column_attrs:
                 if (
-                        schema_event_check and
-                        hasattr(prop.expression, 'info') and
-                        prop.expression.info.get('_ext_mutable_orig_type')
-                        is sqltype
-                ) or (
-                    prop.columns[0].type is sqltype
-                ):
+                    schema_event_check
+                    and hasattr(prop.expression, "info")
+                    and prop.expression.info.get("_ext_mutable_orig_type")
+                    is sqltype
+                ) or (prop.columns[0].type is sqltype):
                     cls.associate_with_attribute(getattr(class_, prop.key))
 
-        event.listen(mapper, 'mapper_configured', listen_for_type)
+        event.listen(mapper, "mapper_configured", listen_for_type)
 
         return sqltype
 
@@ -657,23 +667,29 @@ class MutableComposite(MutableBase):
 
         for parent, key in self._parents.items():
 
-            prop = object_mapper(parent).get_property(key)
+            prop = parent.mapper.get_property(key)
             for value, attr_name in zip(
-                    self.__composite_values__(),
-                    prop._attribute_keys):
-                setattr(parent, attr_name, value)
+                self.__composite_values__(), prop._attribute_keys
+            ):
+                setattr(parent.obj(), attr_name, value)
 
 
 def _setup_composite_listener():
     def _listen_for_type(mapper, class_):
         for prop in mapper.iterate_properties:
-            if (hasattr(prop, 'composite_class') and
-                    isinstance(prop.composite_class, type) and
-                    issubclass(prop.composite_class, MutableComposite)):
+            if (
+                hasattr(prop, "composite_class")
+                and isinstance(prop.composite_class, type)
+                and issubclass(prop.composite_class, MutableComposite)
+            ):
                 prop.composite_class._listen_on_attribute(
-                    getattr(class_, prop.key), False, class_)
+                    getattr(class_, prop.key), False, class_
+                )
+
     if not event.contains(Mapper, "mapper_configured", _listen_for_type):
-        event.listen(Mapper, 'mapper_configured', _listen_for_type)
+        event.listen(Mapper, "mapper_configured", _listen_for_type)
+
+
 _setup_composite_listener()
 
 
@@ -689,11 +705,9 @@ class MutableDict(Mutable, dict):
     solution for the use case of tracking deep changes to a *recursive*
     dictionary structure, such as a JSON structure.  To support this use case,
     build a subclass of  :class:`.MutableDict` that provides appropriate
-    coersion to the values placed in the dictionary so that they too are
+    coercion to the values placed in the dictionary so that they too are
     "mutable", and emit events up to their parent structure.
 
-    .. versionadded:: 0.8
-
     .. seealso::
 
         :class:`.MutableList`
@@ -764,7 +778,7 @@ class MutableList(Mutable, list):
     solution for the use case of tracking deep changes to a *recursive*
     mutable structure, such as a JSON structure.  To support this use case,
     build a subclass of  :class:`.MutableList` that provides appropriate
-    coersion to the values placed in the dictionary so that they too are
+    coercion to the values placed in the dictionary so that they too are
     "mutable", and emit events up to their parent structure.
 
     .. versionadded:: 1.1
@@ -777,6 +791,14 @@ class MutableList(Mutable, list):
 
     """
 
+    def __reduce_ex__(self, proto):
+        return (self.__class__, (list(self),))
+
+    # needed for backwards compatibility with
+    # older pickles
+    def __setstate__(self, state):
+        self[:] = state
+
     def __setitem__(self, index, value):
         """Detect list set events and emit change events."""
         list.__setitem__(self, index, value)
@@ -826,8 +848,8 @@ class MutableList(Mutable, list):
         list.clear(self)
         self.changed()
 
-    def sort(self):
-        list.sort(self)
+    def sort(self, **kw):
+        list.sort(self, **kw)
         self.changed()
 
     def reverse(self):
@@ -844,12 +866,6 @@ class MutableList(Mutable, list):
         else:
             return value
 
-    def __getstate__(self):
-        return list(self)
-
-    def __setstate__(self, state):
-        self[:] = state
-
 
 class MutableSet(Mutable, set):
     """A set type that implements :class:`.Mutable`.
@@ -863,7 +879,7 @@ class MutableSet(Mutable, set):
     solution for the use case of tracking deep changes to a *recursive*
     mutable structure.  To support this use case,
     build a subclass of  :class:`.MutableSet` that provides appropriate
-    coersion to the values placed in the dictionary so that they too are
+    coercion to the values placed in the dictionary so that they too are
     "mutable", and emit events up to their parent structure.
 
     .. versionadded:: 1.1
@@ -947,4 +963,4 @@ class MutableSet(Mutable, set):
         self.update(state)
 
     def __reduce_ex__(self, proto):
-        return (self.__class__, (list(self), ))
+        return (self.__class__, (list(self),))
